#!/usr/bin/env bash
# ------------------------------------------------------------------------------
# This file builds the ecTrans dwarf.
# ------------------------------------------------------------------------------

# Setup required modules.
source load_modules.sh

# Set compilers for make/cmake.
export FC90=ftn
export FC=ftn
export CC=cc
export CXX=CC

# Export used environment variables.
# Note that these are also used by the ecBuild repository.
export BASEDIR=$PWD
export SOURCEDIR=${BASEDIR}/sources
export BUILDDIR=${BASEDIR}/build
export INSTALLDIR=${BASEDIR}/install
export PATH=${PATH}:${INSTALLDIR}/ecbuild/bin/

# Create directories for the build process.
mkdir -p "${BASEDIR}" "${SOURCEDIR}" "${BUILDDIR}" "${INSTALLDIR}"

# Install ecBuild.
cd ${SOURCEDIR}
git clone https://github.com/ecmwf/ecbuild.git
cd ecbuild
git checkout master
# Remove obsolete switch '-Gfast'.
sed -i -e "s/-Gfast//" cmake/compiler_flags/Cray_Fortran.cmake
mkdir -p ${BUILDDIR}/ecbuild
cd ${BUILDDIR}/ecbuild
${SOURCEDIR}/ecbuild/bin/ecbuild --prefix=${INSTALLDIR}/ecbuild ${SOURCEDIR}/ecbuild
make
make install

# Install FIAT.
FIAT_DIR=fiat_cce16_rocm543
cd ${SOURCEDIR}
git clone https://github.com/ecmwf-ifs/fiat
cd fiat
git checkout main
# small fix to include OpenMP in linking of C programs
sed -i -e "s/target_link_libraries( fiat-printbinding OpenMP::OpenMP_C )/target_link_libraries( fiat-printbinding \$\{OpenMP_C_FLAGS\} OpenMP::OpenMP_C )/" src/programs/CMakeLists.txt
rm -rf ${BUILDDIR}/${FIAT_DIR}
mkdir -p ${BUILDDIR}/${FIAT_DIR}
cd ${BUILDDIR}/${FIAT_DIR}
ecbuild -DCMAKE_BUILD_TYPE=RELEASE -DCMAKE_INSTALL_PREFIX=${INSTALLDIR}/${FIAT_DIR} -DENABLE_MPI=ON -DBUILD_SHARED_LIBS=OFF -DENABLE_TESTS=OFF ${SOURCEDIR}/fiat
make -j16
make install

# Install ecTrans
ECTRANS_DIR=ectrans_cce16_rocm543
rm -rf ${BUILDDIR}/${ECTRANS_DIR} ${INSTALLDIR}/${ECTRANS_DIR}
mkdir -p ${BUILDDIR}/${ECTRANS_DIR}
cd ${BUILDDIR}/${ECTRANS_DIR}
ecbuild --prefix=${INSTALLDIR}/${ECTRANS_DIR} -DCMAKE_BUILD_TYPE=RelWithDebInfo -Dfiat_ROOT=${INSTALLDIR}/${FIAT_DIR} -DBUILD_SHARED_LIBS=OFF -DENABLE_FFTW=OFF -DENABLE_GPU=ON -DENABLE_OMPGPU=OFF -DENABLE_ACCGPU=ON -DENABLE_TESTS=OFF -DENABLE_GPU_AWARE_MPI=ON -DENABLE_CPU=ON -DENABLE_ETRANS=ON  -DENABLE_DOUBLE_PRECISION=ON -DENABLE_SINGLE_PRECISION=OFF ${SOURCEDIR}/ectrans
make -j32
make -j32  # run twice because it fails after first time. Some wrong dependencies or something
make install
