MODULE CUDA_GEMM_BATCHED_MOD
  USE HIPBLAS_MOD
  USE PARKIND1, ONLY: JPRD, JPRM, JPIM, JPIB

  IMPLICIT NONE 

!!  PRIVATE
  PUBLIC CUDA_GEMM_BATCHED, CUDA_DGEMM_BATCHED_OVERLOAD, CUDA_DGEMM_BATCHED_1D_OVERLOAD

  INTERFACE CUDA_GEMM_BATCHED
    MODULE PROCEDURE CUDA_DGEMM_BATCHED_OVERLOAD
    MODULE PROCEDURE CUDA_SGEMM_BATCHED_OVERLOAD
    MODULE PROCEDURE CUDA_SGEMM_STRIDED_BATCHED_OVERLOAD
    MODULE PROCEDURE CUDA_DGEMM_BATCHED_1D_OVERLOAD
    MODULE PROCEDURE CUDA_SGEMM_BATCHED_1D_OVERLOAD
    MODULE PROCEDURE CUDA_SGEMM_STRIDED_BATCHED_1D_OVERLOAD
  END INTERFACE CUDA_GEMM_BATCHED

CONTAINS
SUBROUTINE CUDA_DGEMM_BATCHED_OVERLOAD( &
  & TRANSA, TRANSB, &
  & M, N, K, &
  & ALPHA, &
  & AARRAY, LDA, STRIDEA, &
  & BARRAY, LDB, STRIDEB, &
  & BETA, &
  & CARRAY, LDC, STRIDEC, &
  & BATCHCOUNT)
CHARACTER, INTENT(IN) :: TRANSA
CHARACTER, INTENT(IN) :: TRANSB
INTEGER(KIND=JPIM) :: M
INTEGER(KIND=JPIM) :: N
INTEGER(KIND=JPIM) :: K
REAL(KIND=JPRD) :: ALPHA
REAL(KIND=JPRD), DIMENSION(:,:,:) :: AARRAY
INTEGER(KIND=JPIM) :: LDA
INTEGER(KIND=JPIM) :: STRIDEA
REAL(KIND=JPRD), DIMENSION(:,:,:) :: BARRAY
INTEGER(KIND=JPIM) :: LDB
INTEGER(KIND=JPIM) :: STRIDEB
REAL(KIND=JPRD) :: BETA
REAL(KIND=JPRD), DIMENSION(:,:,:) :: CARRAY
INTEGER(KIND=JPIM) :: LDC
INTEGER(KIND=JPIM) :: STRIDEC
INTEGER(KIND=JPIM) :: BATCHCOUNT

CALL HIP_DGEMM_BATCHED( &
  & TRANSA, TRANSB, &
  & M, N, K, &
  & ALPHA, &
  & AARRAY, LDA, STRIDEA, &
  & BARRAY, LDB, STRIDEB, &
  & BETA, &
  & CARRAY, LDC, STRIDEC, &
  & BATCHCOUNT)
END SUBROUTINE CUDA_DGEMM_BATCHED_OVERLOAD

SUBROUTINE CUDA_SGEMM_BATCHED_OVERLOAD( &
  & TRANSA, TRANSB, &
  & M, N, K, &
  & ALPHA, &
  & AARRAY, LDA, STRIDEA, &
  & BARRAY, LDB, STRIDEB, &
  & BETA, &
  & CARRAY, LDC, STRIDEC, &
  & BATCHCOUNT)
CHARACTER, INTENT(IN) :: TRANSA
CHARACTER, INTENT(IN) :: TRANSB
INTEGER(KIND=JPIM) :: M
INTEGER(KIND=JPIM) :: N
INTEGER(KIND=JPIM) :: K
REAL(KIND=JPRM) :: ALPHA
REAL(KIND=JPRM), DIMENSION(:,:,:) :: AARRAY
INTEGER(KIND=JPIM) :: LDA
INTEGER(KIND=JPIM) :: STRIDEA
REAL(KIND=JPRM), DIMENSION(:,:,:) :: BARRAY
INTEGER(KIND=JPIM) :: LDB
INTEGER(KIND=JPIM) :: STRIDEB
REAL(KIND=JPRM) :: BETA
REAL(KIND=JPRM), DIMENSION(:,:,:) :: CARRAY
INTEGER(KIND=JPIM) :: LDC
INTEGER(KIND=JPIM) :: STRIDEC
INTEGER(KIND=JPIM) :: BATCHCOUNT

CALL HIP_SGEMM_BATCHED( &
  & TRANSA, TRANSB, &
  & M, N, K, &
  & ALPHA, &
  & AARRAY, LDA, STRIDEA, &
  & BARRAY, LDB, STRIDEB, &
  & BETA, &
  & CARRAY, LDC, STRIDEC, &
  & BATCHCOUNT)
END SUBROUTINE CUDA_SGEMM_BATCHED_OVERLOAD

SUBROUTINE CUDA_SGEMM_STRIDED_BATCHED_OVERLOAD( &
  & TRANSA, TRANSB, &
  & M, N, K, &
  & ALPHA, &
  & AARRAY, LDA, STRIDEA, &
  & BARRAY, LDB, STRIDEB, &
  & BETA, &
  & CARRAY, LDC, STRIDEC, &
  & BATCHCOUNT)
  CHARACTER, INTENT(IN) :: TRANSA
  CHARACTER, INTENT(IN) :: TRANSB
  INTEGER(KIND=JPIM) :: M
  INTEGER(KIND=JPIM) :: N
  INTEGER(KIND=JPIM) :: K
  REAL(KIND=JPRM) :: ALPHA
  REAL(KIND=JPRM), DIMENSION(:,:,:) :: AARRAY
  INTEGER(KIND=JPIM) :: LDA
  INTEGER(KIND=JPIB) :: STRIDEA
  REAL(KIND=JPRM), DIMENSION(:,:,:) :: BARRAY
  INTEGER(KIND=JPIM) :: LDB
  INTEGER(KIND=JPIB) :: STRIDEB
  REAL(KIND=JPRM) :: BETA
  REAL(KIND=JPRM), DIMENSION(:,:,:) :: CARRAY
  INTEGER(KIND=JPIM) :: LDC
  INTEGER(KIND=JPIB) :: STRIDEC
  INTEGER(KIND=JPIM) :: BATCHCOUNT

  CALL HIP_SGEMM_STRIDED_BATCHED( &
  & TRANSA, TRANSB, &
  & M, N, K, &
  & ALPHA, &
  & AARRAY, LDA, STRIDEA, &
  & BARRAY, LDB, STRIDEB, &
  & BETA, &
  & CARRAY, LDC, STRIDEC, &
  & BATCHCOUNT)
END SUBROUTINE CUDA_SGEMM_STRIDED_BATCHED_OVERLOAD

SUBROUTINE CUDA_DGEMM_BATCHED_1D_OVERLOAD( &
      & TRANSA, TRANSB, &
      & M, N, K, &
      & ALPHA, &
      & AARRAY, LDA, STRIDEA, &
      & BARRAY, LDB, STRIDEB, &
      & BETA, &
      & CARRAY, LDC, STRIDEC, &
      & BATCHCOUNT)
    CHARACTER, INTENT(IN) :: TRANSA
    CHARACTER, INTENT(IN) :: TRANSB
    INTEGER(KIND=JPIM) :: M
    INTEGER(KIND=JPIM) :: N
    INTEGER(KIND=JPIM) :: K
    REAL(KIND=JPRD) :: ALPHA
    REAL(KIND=JPRD), DIMENSION(:) :: AARRAY
    INTEGER(KIND=JPIM) :: LDA
    INTEGER(KIND=JPIM) :: STRIDEA
    REAL(KIND=JPRD), DIMENSION(:,:,:) :: BARRAY
    INTEGER(KIND=JPIM) :: LDB
    INTEGER(KIND=JPIM) :: STRIDEB
    REAL(KIND=JPRD) :: BETA
    REAL(KIND=JPRD), DIMENSION(:) :: CARRAY
    INTEGER(KIND=JPIM) :: LDC
    INTEGER(KIND=JPIM) :: STRIDEC
    INTEGER(KIND=JPIM) :: BATCHCOUNT

    CALL HIP_DGEMM_BATCHED( &
      & TRANSA, TRANSB, &
      & M, N, K, &
      & ALPHA, &
      & AARRAY, LDA, STRIDEA, &
      & BARRAY, LDB, STRIDEB, &
      & BETA, &
      & CARRAY, LDC, STRIDEC, &
      & BATCHCOUNT)
  END SUBROUTINE CUDA_DGEMM_BATCHED_1D_OVERLOAD

  SUBROUTINE CUDA_SGEMM_BATCHED_1D_OVERLOAD( &
      & TRANSA, TRANSB, &
      & M, N, K, &
      & ALPHA, &
      & AARRAY, LDA, STRIDEA, &
      & BARRAY, LDB, STRIDEB, &
      & BETA, &
      & CARRAY, LDC, STRIDEC, &
      & BATCHCOUNT)
    CHARACTER, INTENT(IN) :: TRANSA
    CHARACTER, INTENT(IN) :: TRANSB
    INTEGER(KIND=JPIM) :: M
    INTEGER(KIND=JPIM) :: N
    INTEGER(KIND=JPIM) :: K
    REAL(KIND=JPRM) :: ALPHA
    REAL(KIND=JPRM), DIMENSION(:) :: AARRAY
    INTEGER(KIND=JPIM) :: LDA
    INTEGER(KIND=JPIM) :: STRIDEA
    REAL(KIND=JPRM), DIMENSION(:,:,:) :: BARRAY
    INTEGER(KIND=JPIM) :: LDB
    INTEGER(KIND=JPIM) :: STRIDEB
    REAL(KIND=JPRM) :: BETA
    REAL(KIND=JPRM), DIMENSION(:) :: CARRAY
    INTEGER(KIND=JPIM) :: LDC
    INTEGER(KIND=JPIM) :: STRIDEC
    INTEGER(KIND=JPIM) :: BATCHCOUNT

    CALL HIP_SGEMM_BATCHED( &
      & TRANSA, TRANSB, &
      & M, N, K, &
      & ALPHA, &
      & AARRAY, LDA, STRIDEA, &
      & BARRAY, LDB, STRIDEB, &
      & BETA, &
      & CARRAY, LDC, STRIDEC, &
      & BATCHCOUNT)
  END SUBROUTINE CUDA_SGEMM_BATCHED_1D_OVERLOAD
  
  SUBROUTINE CUDA_SGEMM_STRIDED_BATCHED_1D_OVERLOAD( &
      & TRANSA, TRANSB, &
      & M, N, K, &
      & ALPHA, &
      & AARRAY, LDA, STRIDEA, &
      & BARRAY, LDB, STRIDEB, &
      & BETA, &
      & CARRAY, LDC, STRIDEC, &
      & BATCHCOUNT)
    CHARACTER, INTENT(IN) :: TRANSA
    CHARACTER, INTENT(IN) :: TRANSB
    INTEGER(KIND=JPIM) :: M
    INTEGER(KIND=JPIM) :: N
    INTEGER(KIND=JPIM) :: K
    REAL(KIND=JPRM) :: ALPHA
    REAL(KIND=JPRM), DIMENSION(:) :: AARRAY
    INTEGER(KIND=JPIM) :: LDA
    INTEGER(KIND=JPIB) :: STRIDEA
    REAL(KIND=JPRM), DIMENSION(:,:,:) :: BARRAY
    INTEGER(KIND=JPIM) :: LDB
    INTEGER(KIND=JPIB) :: STRIDEB
    REAL(KIND=JPRM) :: BETA
    REAL(KIND=JPRM), DIMENSION(:) :: CARRAY
    INTEGER(KIND=JPIM) :: LDC
    INTEGER(KIND=JPIB) :: STRIDEC
    INTEGER(KIND=JPIM) :: BATCHCOUNT

    CALL HIP_SGEMM_STRIDED_BATCHED( &
      & TRANSA, TRANSB, &
      & M, N, K, &
      & ALPHA, &
      & AARRAY, LDA, STRIDEA, &
      & BARRAY, LDB, STRIDEB, &
      & BETA, &
      & CARRAY, LDC, STRIDEC, &
      & BATCHCOUNT)
  END SUBROUTINE CUDA_SGEMM_STRIDED_BATCHED_1D_OVERLOAD

END MODULE CUDA_GEMM_BATCHED_MOD
